from typing import TYPE_CHECKING, Dict, List, Optional

from sqlalchemy.sql import text

from fittrackee import db
from fittrackee.utils import decode_short_id
from fittrackee.workouts.exceptions import WorkoutGPXException
from fittrackee.workouts.utils.geometry import (
    get_chart_data_from_segment_points,
)

if TYPE_CHECKING:
    from fittrackee.users.models import User
    from fittrackee.workouts.models import Workout


def get_chart_data(
    workout: "Workout",
    *,
    user: Optional["User"],
    can_see_heart_rate: bool,
    segment_short_id: Optional[str] = None,
) -> Optional[List]:
    """
    Get chart data from segments points if the segments have points, otherwise
    from the gpx file.
    """
    sql = """
        SELECT workout_segments.points
        FROM workout_segments
        WHERE workout_segments.workout_id  = :workout_id"""
    values: Dict = {"workout_id": workout.id}
    if segment_short_id is not None:
        segment_uuid = decode_short_id(segment_short_id)
        sql += """
          AND workout_segments.uuid  = :segment_uuid"""
        values["segment_uuid"] = segment_uuid
    sql += """
        ORDER BY workout_id, start_date"""
    segments_points = db.session.execute(text(sql), values).mappings().all()

    if not segments_points:
        raise WorkoutGPXException(
            "not found",
            (
                f"No segment with id '{segment_short_id}'"
                if segment_short_id is not None
                else "No segments"
            ),
            None,
        )

    if segments_points[0].points:
        return get_chart_data_from_segment_points(
            [segment["points"] for segment in segments_points],
            workout.sport,
            user=user,
            workout_ave_cadence=workout.ave_cadence,
            can_see_heart_rate=can_see_heart_rate,
        )

    return []



import time
from datetime import datetime, timedelta, timezone
from functools import lru_cache, wraps
from typing import Callable, Hashable, Tuple
from uuid import UUID

import nh3
import shortuuid
from sqlalchemy.sql import text

from fittrackee import db


def clean(sql: str, days: int) -> int:
    limit = int(time.time()) - (days * 86400)
    result = db.session.execute(text(sql), {"limit": limit})
    db.session.commit()
    return result.rowcount  # type: ignore


def encode_uuid(uuid_value: UUID) -> str:
    """
    Return short id string from a UUID
    """
    return shortuuid.encode(uuid_value)


def decode_short_id(short_id: str) -> UUID:
    """
    Return UUID from a short id string
    """
    return shortuuid.decode(short_id)


def clean_input(input_text: str, for_markdown_renderer: bool = False) -> str:
    # HTML sanitization
    tags = {"a", "br", "p", "span"}
    attributes = {"a": {"href", "target"}}
    if for_markdown_renderer:
        tags.update({"img", "strong", "em"})
        attributes["img"] = {"src", "alt"}
    return nh3.clean(
        input_text,
        tags=tags,
        attributes=attributes,
    )


class TimedLRUCache:
    def __init__(self, seconds: int, maxsize: int = 128) -> None:
        self.lifetime = timedelta(seconds=seconds)
        self.maxsize = maxsize

    def __call__(self, f: Callable) -> Callable:
        cached_func = lru_cache(maxsize=self.maxsize)(f)
        cached_func.expiration = (  # type: ignore[attr-defined]
            datetime.now(timezone.utc) + self.lifetime
        )

        @wraps(f)
        def wrapper(*args: Tuple, **kwargs: Hashable) -> Callable:
            if datetime.now(timezone.utc) >= cached_func.expiration:  # type: ignore[attr-defined]
                cached_func.cache_clear()
                cached_func.expiration = (  # type: ignore[attr-defined]
                    datetime.now(timezone.utc) + self.lifetime
                )
            return cached_func(*args, **kwargs)

        return wrapper



"""add uuid to notifications table

Revision ID: 70f12f8c0218
Revises: aa7802092404
Create Date: 2024-12-18 14:35:54.901437

"""

from uuid import uuid4

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '70f12f8c0218'
down_revision = 'aa7802092404'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'notifications',
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=True),
    )

    notifications_helper = sa.Table(
        'notifications',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
    )
    connection = op.get_bind()
    for notification in connection.execute(notifications_helper.select()):
        op.execute(f"""
            UPDATE notifications
            SET uuid = '{uuid4()}'
            WHERE notifications.id = {notification.id}
            """)

    op.alter_column('notifications', 'uuid', nullable=False)
    op.create_unique_constraint(
        'notifications_uuid_key', 'notifications', ['uuid']
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('notifications', schema=None) as batch_op:
        batch_op.drop_constraint('notifications_uuid_key', type_='unique')
        batch_op.drop_column('uuid')

    # ### end Alembic commands ###



from fittrackee.exceptions import GenericException


class CommentForbiddenException(GenericException):
    def __init__(self) -> None:
        super().__init__("error", "you do not have permissions")



"""add index on original file and update value

Revision ID: 8221053f0e81
Revises: 45fd75e5f16f
Create Date: 2025-08-17 08:05:45.217576

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '8221053f0e81'
down_revision = '45fd75e5f16f'
branch_labels = None
depends_on = None


def upgrade():
    # For files uploaded before v0.10.0
    op.execute("""
        UPDATE workouts
        SET original_file = gpx
        WHERE workouts.gpx IS NOT NULL
          AND workouts.original_file IS NULL;
    """)

    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_workouts_original_file'), ['original_file'], unique=False)

    # ### end Alembic commands ###



def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_workouts_original_file'))

    # ### end Alembic commands ###



"""remove workout_segment.segment_id and workout.gpx

Revision ID: e63433a1d62e
Revises: 141d3978536a
Create Date: 2025-10-15 12:20:54.297530

"""
from alembic import op
import sqlalchemy as sa
import sys
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'e63433a1d62e'
down_revision = '141d3978536a'
branch_labels = None
depends_on = None


def upgrade():
    # Ensure all workouts are updated before executing upgrade
    connection = op.get_bind()
    workouts_to_update, segments_to_update = connection.execute(
        sa.text("""SELECT
        (SELECT COUNT(*) 
         FROM workouts 
         WHERE workouts.gpx IS NOT NULL AND workouts.start_point_geom IS NULL),
        (SELECT COUNT(*)
         FROM workout_segments
         WHERE workout_segments.geom IS NULL)""")
    ).one()
    if workouts_to_update or segments_to_update:
        print(
            "\nError:\n"
            "Can not run migration, workouts with file and without geometry "
            "exist.\n"
            "Please downgrade to the previous version of FitTrackee (<1.1.0) "
            "and run Workout Refresh CLI with '--add-missing-geometry' option.\n"
        )
        sys.exit(1)

    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workout_segments', schema=None) as batch_op:
        batch_op.alter_column('start_date',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)
        batch_op.drop_column('segment_id')

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_column('gpx')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_column(sa.Column('gpx', sa.VARCHAR(length=255), autoincrement=False, nullable=True))

    with op.batch_alter_table('workout_segments', schema=None) as batch_op:
        batch_op.add_column(sa.Column('segment_id', sa.INTEGER(), autoincrement=False, nullable=True))
        batch_op.alter_column('start_date',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)

    # ### end Alembic commands ###

    op.execute(
        """
        UPDATE workouts
        SET gpx = original_file
        WHERE original_file like '%.gpx';
        """
    )

    for extension in ['fit', 'kml', 'kmz', 'tcx']:
        op.execute(
            f"""
            UPDATE workouts
            SET gpx = REPLACE (original_file, '.{extension}', '.gpx')
            WHERE original_file like '%.{extension}';
            """
        )

    op.execute(
        """
        UPDATE workout_segments
        SET segment_id = ws.calculated_segment_id
        FROM (
                SELECT workout_id,
                       (ROW_NUMBER() OVER (PARTITION BY workout_id ORDER BY start_date) - 1) AS calculated_segment_id
                FROM workout_segments
                ORDER BY workout_id, start_date
              ) ws
        """
    )
    op.alter_column('workout_segments', 'segment_id', nullable=False)



"""add workouts limit for global map and first point geometry to workouts

Revision ID: 0a2e2369b972
Revises: fed271354410
Create Date: 2025-09-17 16:56:55.920185

"""
from alembic import op
import sqlalchemy as sa
from geoalchemy2 import Geometry

# revision identifiers, used by Alembic.
revision = '0a2e2369b972'
down_revision = 'fed271354410'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('app_config', schema=None) as batch_op:
        batch_op.add_column(sa.Column('global_map_workouts_limit', sa.Integer(), server_default='10000', nullable=False))

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_geospatial_column(sa.Column('start_point_geom', Geometry(geometry_type='POINT', srid=4326, dimension=2, spatial_index=False, from_text='ST_GeomFromEWKT', name='geometry'), nullable=True))
        batch_op.create_geospatial_index('idx_workouts_start_point_geom', ['start_point_geom'], unique=False, postgresql_using='gist', postgresql_ops={})

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_geospatial_index('idx_workouts_start_point_geom', postgresql_using='gist', column_name='start_point_geom')
        batch_op.drop_geospatial_column('start_point_geom')

    with op.batch_alter_table('app_config', schema=None) as batch_op:
        batch_op.drop_column('global_map_workouts_limit')

    # ### end Alembic commands ###



import re
from datetime import timedelta
from typing import Optional, Union

import pandas as pd

from ..exceptions import InvalidDurationException


def convert_in_duration(value: str) -> timedelta:
    if not re.match(r"^([0-1]?\d|2[0-3]):[0-5]\d(:[0-5]\d)?$", value):
        raise InvalidDurationException()
    hours = int(value.split(":")[0])
    minutes = int(value.split(":")[1])
    return timedelta(hours=hours, minutes=minutes)


def convert_pace_in_duration(value: str) -> timedelta:
    if not re.match(r"^(\d+):[0-5]\d?$", value):
        raise InvalidDurationException()
    minutes = int(value.split(":")[0])
    seconds = int(value.split(":")[1])
    return timedelta(minutes=minutes, seconds=seconds)


def convert_value_to_integer(
    record_type: str, val: Union[timedelta, float, None]
) -> Optional[int]:
    from ..models import RECORD_TYPES

    if val is None or record_type not in RECORD_TYPES:
        return None

    if isinstance(val, timedelta):  # self.record_type in ["LD", "AP", "BP"]
        return int(val.total_seconds())

    multiplier = (
        100 if record_type in ["AS", "MS"] else 1000  # 'FD' and 'HA
    )
    return round(val * multiplier)


def convert_speed_into_pace_duration(
    speed: Optional[float],
) -> Optional[timedelta]:
    # return pace as duration for 1km
    #
    # note: speed unit is 'km/h'
    if speed is None:
        return None
    if not speed:
        return timedelta(seconds=0)
    return (
        pd.Timedelta(timedelta(minutes=60 / speed)).round("s").to_pytimedelta()
    )


def convert_speed_into_pace_in_sec_per_meter(
    speed: Optional[float],
) -> Optional[float]:
    # return pace in s/m
    #
    # note: speed unit is 'km/h'
    if not speed:
        return None
    return round(1 / (speed / 3.6), 10)



from typing import TYPE_CHECKING, List

import requests

from .base_elevation_service import BaseElevationService
from .exceptions import ElevationServiceException

if TYPE_CHECKING:
    from gpxpy.gpx import GPXTrackPoint


class ValhallaElevationService(BaseElevationService):
    """
    Documentation:
    https://valhalla.github.io/valhalla/api/elevation/api-reference/
    """

    config_key = "VALHALLA_API_URL"
    url_pattern = "{base_url}/height"
    log_label = "Valhalla Elevation API"

    def _get_elevations_for_api(
        self, points: List["GPXTrackPoint"], smooth: bool = False
    ) -> List[int]:
        if not self.url:
            raise ElevationServiceException("Valhalla API: no URL set")

        r = requests.post(
            self.url,
            json={
                "shape": [
                    {"lat": point.latitude, "lon": point.longitude}
                    for point in points
                ]
            },
            timeout=30,
        )
        r.raise_for_status()
        results = r.json().get("height", [])
        return results



"""add whitewater sports

Revision ID: e120ac5a4c5a
Revises: 141d3978536a
Create Date: 2026-01-07 10:06:44.130021

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "e120ac5a4c5a"
down_revision = "d8416412506b"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Canoeing (Whitewater)', True, 1),
               ('Kayaking (Whitewater)', True, 1);
        """
    )


def downgrade():
    op.execute(
        """
        DELETE
        FROM sports
        WHERE label IN ('Canoeing (Whitewater)', 'Kayaking (Whitewater)');
        """
    )



from typing import TYPE_CHECKING, Dict, Optional, Union

import gpxpy.gpx
from gpxpy.gpxfield import parse_time
from lxml import etree as ET

from fittrackee import VERSION
from fittrackee.workouts.exceptions import WorkoutGPXException

from ..constants import NSMAP, TRACK_EXTENSION_NSMAP

if TYPE_CHECKING:
    from fittrackee.workouts.models import Workout


def open_gpx_file(gpx_file: str) -> Optional["gpxpy.gpx.GPX"]:
    gpx_file = open(gpx_file, "r")  # type: ignore
    gpx = gpxpy.parse(gpx_file)
    if len(gpx.tracks) == 0:
        return None
    return gpx


def get_file_extension(filename: str) -> str:
    return filename.rsplit(".", 1)[-1].lower()


VALID_EXTENSIONS = {
    "cadence": "{gpxtpx}cad",
    "heart_rate": "{gpxtpx}hr",
    "power": "{gpxtpx}power",
}


def get_track_extension(calories: Union[int, str]) -> "ET.Element":
    track_extension = ET.Element(
        "{gpxtrkx}TrackStatsExtension",
        nsmap=TRACK_EXTENSION_NSMAP,
    )
    calories_element = ET.SubElement(track_extension, "{gpxtrkx}Calories")
    calories_element.text = str(calories)
    return track_extension


def get_extensions(point: Dict) -> Optional["ET.Element"]:
    track_point_extension = ET.Element("{gpxtpx}TrackPointExtension")
    has_extension = False
    for extension in VALID_EXTENSIONS.keys():
        if point.get(extension) is not None:
            element = ET.SubElement(
                track_point_extension, VALID_EXTENSIONS[extension]
            )
            element.text = str(point[extension])
            has_extension = True
    return track_point_extension if has_extension else None


def generate_gpx(workout: "Workout") -> str:
    if not workout.segments:
        raise WorkoutGPXException("error", "No segments")

    gpx_track = gpxpy.gpx.GPXTrack()
    nsmap = NSMAP

    if workout.calories is not None:
        track_extension = get_track_extension(str(workout.calories))
        gpx_track.extensions.append(track_extension)
        nsmap = {**nsmap, **TRACK_EXTENSION_NSMAP}

    for segment in workout.segments:
        gpx_segment = gpxpy.gpx.GPXTrackSegment()

        for point in segment.points:
            gpx_point = gpxpy.gpx.GPXTrackPoint(
                point.get("latitude"),
                point.get("longitude"),
                elevation=point.get("elevation"),
                time=parse_time(point.get("time", "")),
            )
            extensions = get_extensions(point)
            if extensions is not None:
                gpx_point.extensions.append(extensions)
            gpx_segment.points.append(gpx_point)

        gpx_track.segments.append(gpx_segment)

    gpx = gpxpy.gpx.GPX()
    source = f" (from {workout.source})" if workout.source else ""
    gpx.creator = f"FitTrackee v{VERSION}{source}"
    gpx.nsmap = nsmap
    gpx.tracks.append(gpx_track)

    return gpx.to_xml(prettyprint=True)



import logging

import click

from fittrackee.cli.app import app

from .clean import clean_tokens

logger = logging.getLogger("fittrackee_clean_oauth2_tokens")
logger.setLevel(logging.INFO)


@click.group(name="oauth2")
def oauth2_cli() -> None:
    """Manage OAuth2 tokens."""
    pass


@oauth2_cli.command("clean")
@click.option("--days", type=int, required=True, help="Number of days.")
def clean(
    days: int,
) -> None:
    """Clean tokens expired for more than provided number of days"""
    with app.app_context():
        deleted_rows = clean_tokens(days)
        logger.info(f"Expired deleted tokens: {deleted_rows}.")



"""init social features

Revision ID: aa7802092404
Revises: 4d51a4ca8001
Create Date: 2023-04-13 11:28:53.769936

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from fittrackee.users.roles import UserRole

# revision identifiers, used by Alembic.
revision = 'aa7802092404'
down_revision = '6988082918f8'
branch_labels = None
depends_on = None


visibility_levels = postgresql.ENUM(
    'PUBLIC',
    'FOLLOWERS_AND_REMOTE',  # for a next version, not used for now
    'FOLLOWERS',
    'PRIVATE',
    name='visibility_levels',
)


def upgrade():
    visibility_levels.create(op.get_bind())

    op.create_table(
        'follow_requests',
        sa.Column('follower_user_id', sa.Integer(), nullable=False),
        sa.Column('followed_user_id', sa.Integer(), nullable=False),
        sa.Column('is_approved', sa.Boolean(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ['followed_user_id'],
            ['users.id'],
        ),
        sa.ForeignKeyConstraint(
            ['follower_user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('follower_user_id', 'followed_user_id'),
    )
    op.create_table(
        'comments',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('workout_id', sa.Integer(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('modification_date', sa.DateTime(), nullable=True),
        sa.Column('text', sa.String(), nullable=False),
        sa.Column('suspended_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['workout_id'], ['workouts.id'], ondelete='SET NULL'
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid'),
    )

    op.add_column(
        'comments',
        sa.Column(
            'text_visibility',
            visibility_levels,
            server_default='PRIVATE',
            nullable=False,
        ),
    )

    with op.batch_alter_table('comments', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_comments_user_id'), ['user_id'], unique=False
        )
        batch_op.create_index(
            batch_op.f('ix_comments_workout_id'), ['workout_id'], unique=False
        )

    op.create_table(
        'workout_likes',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('workout_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['workout_id'], ['workouts.id'], ondelete='CASCADE'
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'user_id', 'workout_id', name='user_id_workout_id_unique'
        ),
    )
    op.create_table(
        'comment_likes',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('comment_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['comment_id'], ['comments.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'user_id', 'comment_id', name='user_id_comment_id_unique'
        ),
    )
    op.create_table(
        'mentions',
        sa.Column('comment_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ['comment_id'], ['comments.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('comment_id', 'user_id'),
    )
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                'manually_approves_followers', sa.Boolean(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                'hide_profile_in_users_directory', sa.Boolean(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column(
                'workouts_visibility',
                visibility_levels,
                server_default='PRIVATE',
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column(
                'map_visibility',
                visibility_levels,
                server_default='PRIVATE',
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column('suspended_at', sa.DateTime(), nullable=True)
        )
        batch_op.add_column(sa.Column('role', sa.Integer(), nullable=True))
    user_helper = sa.Table(
        'users',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('admin', sa.String(length=10), nullable=False),
    )
    connection = op.get_bind()
    for user in connection.execute(user_helper.select()):
        op.execute(f"""
            UPDATE users
            SET manually_approves_followers = True,
                hide_profile_in_users_directory = True,
                workouts_visibility = 'PRIVATE',
                map_visibility = 'PRIVATE',
                role = 
                    CASE 
                        WHEN {user.admin} IS TRUE THEN {UserRole.ADMIN.value}
                        ELSE  {UserRole.USER.value}
                    END
            WHERE users.id = {user.id}
            """)
    op.alter_column('users', 'manually_approves_followers', nullable=False)
    op.alter_column('users', 'hide_profile_in_users_directory', nullable=False)
    op.alter_column('users', 'workouts_visibility', nullable=False)
    op.alter_column('users', 'map_visibility', nullable=False)
    op.alter_column('users', 'role', nullable=False)
    op.create_check_constraint(
        'ck_users_role',
        'users',
        f"role IN ({', '.join(UserRole.db_values())})",
    )
    op.drop_column('users', 'admin')

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                'workout_visibility',
                visibility_levels,
                server_default='PRIVATE',
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column(
                'map_visibility',
                visibility_levels,
                server_default='PRIVATE',
                nullable=True,
            )
        )
        batch_op.add_column(
            sa.Column('suspended_at', sa.DateTime(), nullable=True)
        )
    op.execute(
        "UPDATE workouts "
        "SET workout_visibility = 'PRIVATE', "
        "    map_visibility = 'PRIVATE' "
    )
    op.alter_column('workouts', 'workout_visibility', nullable=False)
    op.alter_column('workouts', 'map_visibility', nullable=False)

    op.create_table(
        'notifications',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('from_user_id', sa.Integer(), nullable=True),
        sa.Column('to_user_id', sa.Integer(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('marked_as_read', sa.Boolean(), nullable=False),
        sa.Column('event_object_id', sa.Integer(), nullable=True),
        sa.Column('event_type', sa.String(length=50), nullable=False),
        sa.ForeignKeyConstraint(
            ['from_user_id'], ['users.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(
            ['to_user_id'], ['users.id'], ondelete='CASCADE'
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'from_user_id',
            'to_user_id',
            'event_type',
            'event_object_id',
            name='users_event_unique',
        ),
    )
    with op.batch_alter_table('notifications', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_notifications_from_user_id'),
            ['from_user_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_notifications_to_user_id'),
            ['to_user_id'],
            unique=False,
        )

    op.create_table(
        'blocked_users',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('by_user_id', sa.Integer(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ['by_user_id'], ['users.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'user_id', 'by_user_id', name='blocked_users_unique'
        ),
    )
    with op.batch_alter_table('blocked_users', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_blocked_users_by_user_id'),
            ['by_user_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_blocked_users_user_id'), ['user_id'], unique=False
        )

    op.create_table(
        'reports',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('resolved_at', sa.DateTime(), nullable=True),
        sa.Column('reported_by', sa.Integer(), nullable=True),
        sa.Column('reported_comment_id', sa.Integer(), nullable=True),
        sa.Column('reported_user_id', sa.Integer(), nullable=True),
        sa.Column('reported_workout_id', sa.Integer(), nullable=True),
        sa.Column('resolved_by', sa.Integer(), nullable=True),
        sa.Column('resolved', sa.Boolean(), nullable=False),
        sa.Column('object_type', sa.String(length=50), nullable=False),
        sa.Column('note', sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ['reported_by'], ['users.id'], ondelete='SET NULL'
        ),
        sa.ForeignKeyConstraint(
            ['reported_comment_id'], ['comments.id'], ondelete='SET NULL'
        ),
        sa.ForeignKeyConstraint(
            ['reported_user_id'], ['users.id'], ondelete='SET NULL'
        ),
        sa.ForeignKeyConstraint(
            ['reported_workout_id'], ['workouts.id'], ondelete='SET NULL'
        ),
        sa.ForeignKeyConstraint(
            ['resolved_by'], ['users.id'], ondelete='SET NULL'
        ),
        sa.PrimaryKeyConstraint('id'),
    )
    with op.batch_alter_table('reports', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_reports_reported_by'), ['reported_by'], unique=False
        )
        batch_op.create_index(
            batch_op.f('ix_reports_reported_comment_id'),
            ['reported_comment_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_reports_reported_user_id'),
            ['reported_user_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_reports_reported_workout_id'),
            ['reported_workout_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_reports_resolved_by'), ['resolved_by'], unique=False
        )
        batch_op.create_index(
            batch_op.f('ix_reports_object_type'), ['object_type'], unique=False
        )

    op.create_table(
        'report_comments',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('report_id', sa.Integer(), nullable=True),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('comment', sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ['report_id'], ['reports.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
    )
    with op.batch_alter_table('report_comments', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_report_comments_report_id'),
            ['report_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_report_comments_user_id'), ['user_id'], unique=False
        )

    with op.batch_alter_table('records', schema=None) as batch_op:
        batch_op.create_index(
            'workout_records', ['workout_id', 'record_type'], unique=False
        )

    op.create_table(
        'report_actions',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('moderator_id', sa.Integer(), nullable=True),
        sa.Column('report_id', sa.Integer(), nullable=False),
        sa.Column('comment_id', sa.Integer(), nullable=True),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('workout_id', sa.Integer(), nullable=True),
        sa.Column('action_type', sa.String(length=50), nullable=False),
        sa.Column('reason', sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ['moderator_id'], ['users.id'], ondelete='SET NULL'
        ),
        sa.ForeignKeyConstraint(
            ['report_id'], ['reports.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(
            ['comment_id'], ['comments.id'], ondelete='SET NULL'
        ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['workout_id'], ['workouts.id'], ondelete='SET NULL'
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid'),
    )
    with op.batch_alter_table('report_actions', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_report_actions_moderator_id'),
            ['moderator_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_report_actions_report_id'),
            ['report_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_report_actions_comment_id'),
            ['comment_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_report_actions_user_id'), ['user_id'], unique=False
        )
        batch_op.create_index(
            batch_op.f('ix_report_actions_workout_id'),
            ['workout_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_report_actions_created_at'),
            ['created_at'],
            unique=False,
        )

    op.create_table(
        'report_action_appeals',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column('action_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('moderator_id', sa.Integer(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('approved', sa.Boolean(), nullable=True),
        sa.Column('text', sa.String(), nullable=False),
        sa.Column('reason', sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ['action_id'], ['report_actions.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(
            ['moderator_id'], ['users.id'], ondelete='SET NULL'
        ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'action_id', 'user_id', name='action_id_user_id_unique'
        ),
        sa.UniqueConstraint('uuid'),
    )
    with op.batch_alter_table('report_action_appeals', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_report_action_appeals_action_id'),
            ['action_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_report_action_appeals_moderator_id'),
            ['moderator_id'],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f('ix_report_action_appeals_user_id'),
            ['user_id'],
            unique=False,
        )


def downgrade():
    with op.batch_alter_table('report_action_appeals', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_report_action_appeals_user_id'))
        batch_op.drop_index(
            batch_op.f('ix_report_action_appeals_moderator_id')
        )
        batch_op.drop_index(batch_op.f('ix_report_action_appeals_action_id'))

    op.drop_table('report_action_appeals')

    with op.batch_alter_table('report_actions', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_report_actions_created_at'))
        batch_op.drop_index(batch_op.f('ix_report_actions_user_id'))
        batch_op.drop_index(batch_op.f('ix_report_actions_report_id'))
        batch_op.drop_index(batch_op.f('ix_report_actions_moderator_id'))

    op.drop_table('report_actions')

    with op.batch_alter_table('records', schema=None) as batch_op:
        batch_op.drop_index('workout_records')

    with op.batch_alter_table('report_comments', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_report_comments_user_id'))
        batch_op.drop_index(batch_op.f('ix_report_comments_report_id'))

    op.drop_table('report_comments')
    with op.batch_alter_table('reports', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_reports_object_type'))
        batch_op.drop_index(batch_op.f('ix_reports_resolved_by'))
        batch_op.drop_index(batch_op.f('ix_reports_reported_workout_id'))
        batch_op.drop_index(batch_op.f('ix_reports_reported_user_id'))
        batch_op.drop_index(batch_op.f('ix_reports_reported_comment_id'))
        batch_op.drop_index(batch_op.f('ix_reports_reported_by'))

    op.drop_table('reports')

    with op.batch_alter_table('blocked_users', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_blocked_users_user_id'))
        batch_op.drop_index(batch_op.f('ix_blocked_users_by_user_id'))

    op.drop_table('blocked_users')

    with op.batch_alter_table('notifications', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_notifications_to_user_id'))
        batch_op.drop_index(batch_op.f('ix_notifications_from_user_id'))
    op.drop_table('notifications')

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_column('map_visibility')
        batch_op.drop_column('workout_visibility')
        batch_op.drop_column('suspended_at')

    op.add_column('users', sa.Column('admin', sa.Boolean(), nullable=True))
    batch_op.drop_constraint(batch_op.f('ck_users_role'))

    user_helper = sa.Table(
        'users',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('role', sa.Integer(), nullable=False),
    )
    connection = op.get_bind()
    for user in connection.execute(user_helper.select()):
        op.execute(f"""
            UPDATE users
            SET admin = {user.role} <> {UserRole.USER.value}
            WHERE users.id = {user.id}
            """)

    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.alter_column('admin', nullable=False)
        batch_op.drop_column('role')
        batch_op.drop_column('suspended_at')
        batch_op.drop_column('map_visibility')
        batch_op.drop_column('workouts_visibility')
        batch_op.drop_column('hide_profile_in_users_directory')
        batch_op.drop_column('manually_approves_followers')

    op.drop_table('mentions')
    op.drop_table('comment_likes')
    op.drop_table('workout_likes')
    with op.batch_alter_table('comments', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_comments_workout_id'))
        batch_op.drop_index(batch_op.f('ix_comments_user_id'))

    op.drop_table('comments')
    op.drop_table('follow_requests')
    visibility_levels.drop(op.get_bind())



"""change split_workout_charts default value

Revision ID: ff53544547cd
Revises: e120ac5a4c5a
Create Date: 2026-01-21 09:07:29.233113

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ff53544547cd'
down_revision = 'e120ac5a4c5a'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('users', "split_workout_charts", server_default="true")


def downgrade():
    op.alter_column('users', "split_workout_charts", server_default="false")



from typing import Dict

from fittrackee.templates import I18nTemplate


class FeedItemTemplate(I18nTemplate):
    def get_item_data(self, template: str, lang: str, data: Dict) -> Dict:
        parts = ["title.txt", "body.html"]
        return self.get_all_contents(template, lang, parts, data)



import re


def is_valid_email(email: str) -> bool:
    """
    Return if email format is valid
    """
    if not email:
        return False
    mail_pattern = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
    return re.match(mail_pattern, email) is not None


def check_password(password: str) -> str:
    """
    Verify if password have more than 8 characters
    If not, it returns error message
    """
    if len(password) < 8:
        return "password: 8 characters required\n"
    return ""


def check_username(username: str) -> str:
    """
    Return if username is valid
    If not, it returns error messages
    """
    ret = ""
    if not 2 < len(username) < 31:
        ret += "username: 3 to 30 characters required\n"
    if not re.match(r"^[a-zA-Z0-9_]+$", username):
        ret += (
            "username: only alphanumeric characters and the "
            'underscore character "_" allowed\n'
        )
    return ret


def register_controls(username: str, email: str, password: str) -> str:
    """
    Verify if username, email and passwords are valid
    If not, it returns error messages
    """
    ret = check_username(username)
    if not is_valid_email(email):
        ret += "email: valid email must be provided\n"
    ret += check_password(password)
    return ret



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from datetime import datetime

    from ..workout_from_file.workout_point import WorkoutPoint


class BaseWeather(ABC):
    def __init__(self, api_key: str) -> None:
        self.api_key: str = api_key

    @abstractmethod
    def _get_data(
        self, latitude: float, longitude: float, time: "datetime"
    ) -> Optional[Dict]:
        """
        Expected dict:
        {
            "humidity": 0.69,
            "icon": "partly-cloudy-day",
            "temperature": 12.26,
            "wind": 3.49,
            "windBearing": 315
        }

        FitTrackee expects the following units:
          temperature: Celsius,
          humidity: in fraction (rather than percent)
          windSpeed: m/s
          windBearing: direction wind is from in degrees (0 is north)

        Expected icon values (for UI):
        - "clear-day",
        - "clear-night",
        - "cloudy",
        - "fog",
        - "partly-cloudy-day",
        - "partly-cloudy-night",
        - "rain",
        - "sleet",
        - "snow",
        - "wind"
        """
        pass

    def get_weather(self, point: "WorkoutPoint") -> Optional[Dict]:
        if not point.time:
            # if there's no time associated with the point,
            # we cannot get weather
            return None

        return self._get_data(point.latitude, point.longitude, point.time)



from flask_babel import gettext as _

# for strings detection by gettext
AVAILABLE_SPORTS = [
    _("Canoeing"),
    _("Canoeing (Whitewater)"),
    _("Cycling (Sport)"),
    _("Cycling (Transport)"),
    _("Cycling (Trekking)"),
    _("Cycling (Virtual)"),
    _("Halfbike"),
    _("Hiking"),
    _("Kayaking"),
    _("Ice Skating"),
    _("Inline Skating"),
    _("Kayaking (Whitewater)"),
    _("Mountain Biking"),
    _("Mountain Biking (Electric)"),
    _("Mountaineering"),
    _("Open Water Swimming"),
    _("Padel (Outdoor)"),
    _("Paragliding"),
    _("Rowing"),
    _("Running"),
    _("Skiing (Alpine)"),
    _("Skiing (Cross Country)"),
    _("Snowshoes"),
    _("Standup Paddleboarding"),
    _("Swimrun"),
    _("Tennis (Outdoor)"),
    _("Trail"),
    _("Walking"),
    _("Windsurfing"),
]



from typing import TYPE_CHECKING, List, Tuple, Union

from fittrackee.constants import ElevationDataSource

from .open_elevation_service import OpenElevationService
from .valhalla_elevation_service import ValhallaElevationService

if TYPE_CHECKING:
    from gpxpy.gpx import GPXTrackPoint

WINDOW_LEN = 51


class ElevationService:
    """
    Available elevation services:
    - Open Elevation (with or without smoothing processing)
    - Valhalla
    """

    def __init__(self, elevation_data_source: "ElevationDataSource") -> None:
        self.elevation_service, self.smooth, self.elevation_data_source = (
            self._get_elevation_service(elevation_data_source)
        )

    @staticmethod
    def _get_elevation_service(
        elevation_data_source: "ElevationDataSource",
    ) -> Tuple[
        Union["OpenElevationService", "ValhallaElevationService", None],
        bool,
        "ElevationDataSource",
    ]:
        if elevation_data_source == ElevationDataSource.FILE:
            return None, False, elevation_data_source

        service: Union[
            "OpenElevationService", "ValhallaElevationService", None
        ] = None
        if elevation_data_source in [
            ElevationDataSource.OPEN_ELEVATION,
            ElevationDataSource.OPEN_ELEVATION_SMOOTH,
        ]:
            service = OpenElevationService()

        if elevation_data_source == ElevationDataSource.VALHALLA:
            service = ValhallaElevationService()

        if service and service.is_enabled:
            return (
                service,
                (
                    elevation_data_source
                    == ElevationDataSource.OPEN_ELEVATION_SMOOTH
                ),
                elevation_data_source,
            )
        return None, False, ElevationDataSource.FILE

    def get_elevations(self, points: List["GPXTrackPoint"]) -> List[int]:
        if not self.elevation_service:
            return []

        return self.elevation_service.get_elevations(
            points, smooth=self.smooth
        )



from abc import ABC, abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Tuple

from sqlalchemy import asc, desc

from fittrackee.users.models import User
from fittrackee.workouts.models import Workout

if TYPE_CHECKING:
    from logging import Logger


class AbstractWorkoutsRefreshService(ABC):
    def __init__(
        self,
        logger: "Logger",
        per_page: Optional[int] = 10,
        page: Optional[int] = 1,
        order: Optional[str] = "asc",
        user: Optional[str] = None,
        sport_id: Optional[int] = None,
        new_sport_id: Optional[int] = None,
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        verbose: bool = False,
    ) -> None:
        self.per_page: Optional[int] = per_page
        self.page: Optional[int] = page
        self.order: Optional[str] = order
        self.username: Optional[str] = user
        self.sport_id: Optional[int] = sport_id
        self.new_sport_id: Optional[int] = new_sport_id
        self.date_from: Optional["datetime"] = date_from
        self.date_to: Optional["datetime"] = date_to
        self.logger: "Logger" = logger
        self.verbose: bool = verbose

    def _log_if_verbose(self, message: str) -> None:
        if not self.verbose:
            return
        self.logger.info(message)

    def _get_workouts(self, filters: list) -> Tuple[List["Workout"], int]:
        workouts_to_refresh_query = Workout.query
        if self.username:
            workouts_to_refresh_query = workouts_to_refresh_query.join(
                User, User.id == Workout.user_id
            )
            filters.extend([User.username == self.username])
        if self.sport_id:
            filters.extend([Workout.sport_id == self.sport_id])
        if self.date_from:
            filters.extend([Workout.workout_date >= self.date_from])
        if self.date_to:
            filters.extend([Workout.workout_date <= self.date_to])

        workouts_to_refresh = (
            workouts_to_refresh_query.filter(*filters)
            .order_by(
                asc(Workout.workout_date)
                if self.order == "asc"
                else desc(Workout.workout_date)
            )
            .paginate(page=self.page, per_page=self.per_page, error_out=False)
            .items
        )

        total = len(workouts_to_refresh)
        if not total:
            self.logger.info("No workouts to refresh.")
            return [], 0

        self.logger.info(f"Number of workouts to refresh: {total}")
        return workouts_to_refresh, total

    @abstractmethod
    def refresh(self) -> int:
        pass



import logging
from logging.config import fileConfig

from alembic import context
from flask import current_app
from geoalchemy2 import alembic_helpers

from fittrackee.database import TZDateTime

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)
logger = logging.getLogger("alembic.env")


def get_engine():
    return current_app.extensions["migrate"].db.get_engine()


def get_engine_url():
    try:
        return (
            get_engine()
            .url.render_as_string(hide_password=False)
            .replace("%", "%%")
        )
    except AttributeError:
        return str(get_engine().url).replace("%", "%%")


# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
config.set_main_option("sqlalchemy.url", get_engine_url())
target_db = current_app.extensions["migrate"].db

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def get_metadata():
    if hasattr(target_db, "metadatas"):
        return target_db.metadatas[None]
    return target_db.metadata


def render_item(obj_type, obj, autogen_context):
    """Apply custom rendering for selected items."""
    spatial_type = alembic_helpers.render_item(obj_type, obj, autogen_context)
    if spatial_type:
        return spatial_type

    # For the custom type
    if obj_type == "type" and isinstance(obj, TZDateTime):
        import_name = obj.__class__.__name__
        autogen_context.imports.add(
            f"from fittrackee.database import {import_name}"
        )
        return "%r" % obj

    # default rendering for other objects
    return False


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=get_metadata(),
        literal_binds=True,
        include_object=alembic_helpers.include_object,
        process_revision_directives=alembic_helpers.writer,
        render_item=render_item,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    # this callback is used to prevent an auto-migration from being generated
    # when there are no changes to the schema
    # reference: http://alembic.zzzcomputing.com/en/latest/cookbook.html
    def process_revision_directives(context, revision, directives):
        if getattr(config.cmd_opts, "autogenerate", False):
            script = directives[0]
            if script.upgrade_ops.is_empty():
                directives[:] = []
                logger.info("No changes in schema detected.")

    conf_args = current_app.extensions["migrate"].configure_args
    if conf_args.get("process_revision_directives") is None:
        conf_args["process_revision_directives"] = alembic_helpers.writer

    connectable = get_engine()

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=get_metadata(),
            include_object=alembic_helpers.include_object,
            render_item=render_item,
            **conf_args,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



USER_DATE_FORMAT = "MM/dd/yyyy"
USER_LANGUAGE = "en"
USER_TIMEZONE = "Europe/Paris"
USER_LINK_TEMPLATE = (
    '<a href="{profile_url}" target="_blank" rel="noopener noreferrer">'
    "{username}</a>"
)

ADMINISTRATOR_NOTIFICATION_TYPES = [
    "account_creation",
]
MODERATOR_NOTIFICATION_TYPES = [
    "report",
    "suspension_appeal",
    "user_warning_appeal",
]
NOTIFICATION_TYPES = (
    ADMINISTRATOR_NOTIFICATION_TYPES
    + MODERATOR_NOTIFICATION_TYPES
    + [
        "comment_like",
        "comment_suspension",
        "comment_unsuspension",
        "follow",
        "follow_request",
        "follow_request_approved",
        "mention",
        "user_data_export",
        "user_warning",
        "user_warning_lifting",
        "workouts_archive_upload",
        "workout_comment",
        "workout_like",
        "workout_suspension",
        "workout_unsuspension",
    ]
)

NOTIFICATIONS_PREFERENCES_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "Notifications",
    "description": "Preferences for UI notifications",
    "type": "object",
    "properties": {
        "account_creation": {
            "type": "boolean",
            "default": True,
        },
        "comment_like": {
            "type": "boolean",
            "default": True,
        },
        "follow": {
            "type": "boolean",
            "default": True,
        },
        "follow_request": {
            "type": "boolean",
            "default": True,
        },
        "follow_request_approved": {
            "type": "boolean",
            "default": True,
        },
        "mention": {
            "type": "boolean",
            "default": True,
        },
        "workout_comment": {
            "type": "boolean",
            "default": True,
        },
        "workout_like": {
            "type": "boolean",
            "default": True,
        },
    },
    "additionalProperties": False,
}

MESSAGE_PREFERENCES_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "Messages",
    "description": "Preferences for messages displayed on UI",
    "type": "object",
    "properties": {
        "warning_about_large_number_of_workouts_on_map": {
            "type": "boolean",
            "default": True,
        },
    },
    "additionalProperties": False,
}



"""add cycling (trekking) sport

Revision ID: 4d51a4ca8001
Revises: 14f48e46f320
Create Date: 2023-12-20 13:45:48.654139

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4d51a4ca8001'
down_revision = '14f48e46f320'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Cycling (Trekking)', True, 1)
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Cycling (Trekking)';
        """
    )



from typing import Dict, Union

from flask import Blueprint, request
from sqlalchemy import and_, or_

from fittrackee.oauth2.server import require_auth
from fittrackee.responses import HttpResponse, handle_error_and_return_response
from fittrackee.users.models import User
from fittrackee.visibility_levels import VisibilityLevel

from .models import Workout

timeline_blueprint = Blueprint("timeline", __name__)

DEFAULT_WORKOUTS_PER_PAGE = 5


@timeline_blueprint.route("/timeline", methods=["GET"])
@require_auth(scopes=["workouts:read"])
def get_user_timeline(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Get workouts visible to authenticated user.

    **Scope**: ``workouts:read``

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /api/timeline HTTP/1.1

    - with some query parameters:

    .. sourcecode:: http

      GET /api/timeline?page=2  HTTP/1.1

    **Example responses**:

    - returning at least one workout:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "workouts": [
              {
                "ascent": null,
                "ave_cadence": null,
                "ave_hr": null,
                "ave_pace": null,
                "ave_power": null,
                "ave_speed": 10.0,
                "best_pace": null,
                "bounds": [],
                "creation_date": "Sun, 14 Jul 2019 13:51:01 GMT",
                "descent": null,
                "description": null,
                "distance": 10.0,
                "duration": "0:17:04",
                "equipments": [],
                "id": "kjxavSTUrJvoAh2wvCeGEF",
                "map": null,
                "max_alt": null,
                "max_cadence": null,
                "max_hr": null,
                "max_power": null,
                "max_speed": 10.0,
                "min_alt": null,
                "modification_date": null,
                "moving": "0:17:04",
                "next_workout": 3,
                "notes": null,
                "pauses": null,
                "previous_workout": null,
                "records": [
                  {
                    "id": 4,
                    "record_type": "MS",
                    "sport_id": 1,
                    "user": "admin",
                    "value": 10.0,
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  },
                  {
                    "id": 13,
                    "record_type": "HA",
                    "sport_id": 1,
                    "user": "Sam",
                    "value": 43.97,
                    "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                    "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                  },
                  {
                    "id": 3,
                    "record_type": "LD",
                    "sport_id": 1,
                    "user": "admin",
                    "value": "0:17:04",
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  },
                  {
                    "id": 2,
                    "record_type": "FD",
                    "sport_id": 1,
                    "user": "admin",
                    "value": 10.0,
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  },
                  {
                    "id": 1,
                    "record_type": "AS",
                    "sport_id": 1,
                    "user": "admin",
                    "value": 10.0,
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  }
                ],
                "segments": [],
                "sport_id": 1,
                "title": null,
                "user": "admin",
                "weather_end": null,
                "weather_start": null,
                "with_file": false,
                "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT"
              }
            ]
          },
          "status": "success"
        }

    - returning no workouts

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
            "data": {
                "workouts": []
            },
            "status": "success"
        }

    :query integer page: page if using pagination (default: 1)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    try:
        params = request.args.copy()
        page = int(params.get("page", 1))
        following_ids = auth_user.get_following_user_ids()
        blocked_users = auth_user.get_blocked_user_ids()
        blocked_by_users = auth_user.get_blocked_by_user_ids()
        workouts_pagination = (
            Workout.query.join(
                User,
                Workout.user_id == User.id,
            )
            .filter(
                or_(
                    # get all authenticated user workouts
                    Workout.user_id == auth_user.id,
                    # gel followed users workouts, that are not suspended
                    # and user is not blocked
                    and_(
                        Workout.suspended_at == None,  # noqa
                        and_(
                            Workout.user_id.in_(following_ids),
                            Workout.user_id.not_in(
                                blocked_users + blocked_by_users
                            ),
                            Workout.workout_visibility.in_(
                                [
                                    VisibilityLevel.FOLLOWERS,
                                    VisibilityLevel.PUBLIC,
                                ]
                            ),
                        ),
                    ),
                ),
                User.suspended_at == None,  # noqa
            )
            .order_by(
                Workout.workout_date.desc(),
            )
            .paginate(
                page=page, per_page=DEFAULT_WORKOUTS_PER_PAGE, error_out=False
            )
        )
        workouts = workouts_pagination.items
        return {
            "status": "success",
            "data": {
                "workouts": [
                    workout.serialize(user=auth_user) for workout in workouts
                ]
            },
            "pagination": {
                "has_next": workouts_pagination.has_next,
                "has_prev": workouts_pagination.has_prev,
                "page": workouts_pagination.page,
                "pages": workouts_pagination.pages,
                "total": workouts_pagination.total,
            },
        }
    except Exception as e:
        return handle_error_and_return_response(e)



from typing import List, Optional, Union

from fittrackee.equipments.models import Equipment
from fittrackee.users.models import User
from fittrackee.utils import decode_short_id
from fittrackee.workouts.models import Sport

from .exceptions import InvalidEquipmentException, InvalidEquipmentsException

SPORT_EQUIPMENT_TYPES = {
    "Bike": [
        "Cycling (Sport)",
        "Cycling (Transport)",
        "Cycling (Trekking)",
        "Halfbike",
        "Mountain Biking",
        "Mountain Biking (Electric)",
    ],
    "Bike Trainer": ["Cycling (Virtual)"],
    "Board": ["Standup Paddleboarding", "Windsurfing"],
    "Kayak_Boat": [
        "Canoeing",
        "Canoeing (Whitewater)",
        "Kayaking",
        "Kayaking (Whitewater)",
        "Rowing",
    ],
    "Skates": ["Ice Skating", "Inline Skating"],
    "Paddle": [
        "Canoeing",
        "Canoeing (Whitewater)",
        "Kayaking",
        "Kayaking (Whitewater)",
        "Rowing",
        "Standup Paddleboarding",
    ],
    "Racket": ["Padel (Outdoor)", "Tennis (Outdoor)"],
    "Skis": ["Skiing (Alpine)", "Skiing (Cross Country)"],
    "Shoes": [
        "Hiking",
        "Mountaineering",
        "Padel (Outdoor)",
        "Running",
        "Swimrun",
        "Tennis (Outdoor)",
        "Trail",
        "Walking",
        "Cycling (Sport)",
        "Cycling (Transport)",
        "Cycling (Trekking)",
        "Halfbike",
        "Mountain Biking",
        "Mountain Biking (Electric)",
    ],
    "Snowshoes": ["Snowshoes"],
}
MAX_MISC_LIMIT = 5
TYPE_ERROR_MESSAGE = "equipment_ids must be an array of strings"


def handle_pieces_of_equipment(
    equipment_short_ids: Optional[List[str]],
    auth_user: User,
    sport_id: int,
    existing_equipments: Optional[List[Equipment]] = None,
) -> Union[List[Equipment], None]:
    if equipment_short_ids is None:
        return equipment_short_ids

    equipments_list: List[Equipment] = []
    if not equipment_short_ids:
        return equipments_list

    if not isinstance(equipment_short_ids, list):
        raise InvalidEquipmentsException(TYPE_ERROR_MESSAGE)

    sport = Sport.query.filter_by(id=sport_id).first()
    if not sport:
        raise InvalidEquipmentsException(f"sport id {sport_id} not found")

    equipment_types = []
    misc_count = 0
    for equipment_short_id in equipment_short_ids:
        if not isinstance(equipment_short_id, str):
            raise InvalidEquipmentsException(TYPE_ERROR_MESSAGE)

        equipment = Equipment.query.filter_by(
            uuid=decode_short_id(equipment_short_id), user_id=auth_user.id
        ).first()
        if not equipment:
            raise InvalidEquipmentException(
                status="not_found",
                message=(
                    f"equipment with id {equipment_short_id} does not exist"
                ),
                equipment_short_id=equipment_short_id,
            )

        if not equipment.is_active and (
            not existing_equipments or equipment not in existing_equipments
        ):
            raise InvalidEquipmentException(
                status="inactive",
                message=(
                    f"equipment with id {equipment_short_id} is inactive"
                ),
                equipment_short_id=equipment_short_id,
            )

        if equipment.equipment_type.label == "Misc":
            misc_count += 1
            if misc_count > MAX_MISC_LIMIT:
                raise InvalidEquipmentsException(
                    f"a maximum of {MAX_MISC_LIMIT} pieces of Misc equipment "
                    "can be added"
                )
        else:
            if sport.label not in SPORT_EQUIPMENT_TYPES.get(
                equipment.equipment_type.label, []
            ):
                raise InvalidEquipmentException(
                    status="invalid",
                    message=(
                        f"invalid equipment id {equipment.short_id} "
                        f"for sport {sport.label}"
                    ),
                    equipment_short_id=equipment_short_id,
                )

            if equipment.equipment_type in equipment_types:
                raise InvalidEquipmentsException(
                    "only one piece of equipment per type can be provided"
                )
            equipment_types.append(equipment.equipment_type)

        equipments_list.append(equipment)

    return equipments_list



from dataclasses import dataclass
from datetime import timedelta
from decimal import Decimal
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from fittrackee.constants import PaceSpeedDisplay
from fittrackee.workouts.constants import (
    CADENCE_SPORTS,
    POWER_SPORTS,
    RPM_CADENCE_SPORTS,
    SPM_CADENCE_SPORTS,
    SPORTS_WITHOUT_ELEVATION_DATA,
)

if TYPE_CHECKING:
    from fittrackee.users.models import User
    from fittrackee.workouts.models import Sport


@dataclass
class SportDisplayedData:
    display_elevation: bool
    display_pace: bool
    display_power: bool
    display_speed: bool
    display_cadence: bool
    display_spm_cadence: bool
    display_rpm_cadence: bool


def get_sports_displayed_data(
    sports: List["Sport"],
    user: Optional["User"],
    force_display_speed: bool = False,
) -> Dict:
    """
    Get sports display config from:
    - 'user_sports_preferences' if user is provided and a preference exists
    - otherwise from 'sports'

    'force_display_speed' allows to display speed in Workout list regardless
    sport preferences
    """
    from fittrackee.users.models import UserSportPreference

    sports_data_visibility = {}
    for sport in sports:
        sport_preferences = (
            UserSportPreference.query.filter_by(
                sport_id=sport.id, user_id=user.id
            ).first()
            if user
            else None
        )
        pace_speed_display = (
            sport_preferences.pace_speed_display
            if sport_preferences
            else sport.pace_speed_display
        )
        sports_data_visibility[sport.id] = SportDisplayedData(
            display_elevation=sport.label not in SPORTS_WITHOUT_ELEVATION_DATA,
            display_pace=pace_speed_display != PaceSpeedDisplay.SPEED,
            display_speed=force_display_speed
            or pace_speed_display != PaceSpeedDisplay.PACE,
            display_power=sport.label in POWER_SPORTS,
            display_cadence=sport.label in CADENCE_SPORTS,
            display_spm_cadence=sport.label in SPM_CADENCE_SPORTS,
            display_rpm_cadence=sport.label in RPM_CADENCE_SPORTS,
        )
    return sports_data_visibility


def get_sport_displayed_data(
    sport: "Sport", user: Optional["User"], force_display_speed: bool = False
) -> "SportDisplayedData":
    return get_sports_displayed_data([sport], user, force_display_speed)[
        sport.id
    ]


def get_pace(
    pace: Optional[timedelta], sport_data_visibility: "SportDisplayedData"
) -> Optional[str]:
    if pace is None or not sport_data_visibility.display_pace:
        return None

    return str(pace)


def get_speed(
    value: Union[float, Decimal, None],
    sport_data_visibility: "SportDisplayedData",
    pace: Optional[timedelta] = None,
) -> Optional[float]:
    if value is None:
        return None

    if (
        # in case pace is not yet calculated for workouts created
        # before 1.1.0
        sport_data_visibility.display_pace and pace is None
    ) or sport_data_visibility.display_speed:
        return float(value)

    return None


def get_elevation_data(
    value: Optional[float],
    can_see_analysis_data: bool,
    sport_data_visibility: "SportDisplayedData",
) -> Optional[float]:
    if (
        value is None
        or not can_see_analysis_data
        or not sport_data_visibility.display_elevation
    ):
        return None

    return float(value)


def get_cadence(
    cadence: Optional[int], sport_data_visibility: "SportDisplayedData"
) -> Optional[int]:
    if cadence is None:
        return None

    if sport_data_visibility.display_rpm_cadence:
        return cadence

    if sport_data_visibility.display_spm_cadence:
        return cadence * 2

    return None


def get_power(
    power: Optional[int], sport_data_visibility: "SportDisplayedData"
) -> Optional[int]:
    if power is None or not sport_data_visibility.display_power:
        return None

    return power



"""add imperial units preferences

Revision ID: 07188ca7620a
Revises: 080acc8ee956
Create Date: 2021-11-13 19:11:17.753567

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '07188ca7620a'
down_revision = '080acc8ee956'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'users',
        sa.Column('imperial_units', sa.Boolean(), nullable=True),
    )
    op.execute("UPDATE users SET imperial_units = false")
    op.alter_column('users', 'imperial_units', nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'imperial_units')
    # ### end Alembic commands ###



from typing import Dict, Union

from flask import Blueprint, request
from sqlalchemy import and_, asc, desc, exc, or_

from fittrackee import db
from fittrackee.comments.models import Comment
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    HttpResponse,
    NotFoundErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.utils import decode_short_id
from fittrackee.visibility_levels import VisibilityLevel

from .models import Notification, User

notifications_blueprint = Blueprint("notifications", __name__)


DEFAULT_NOTIFICATION_PER_PAGE = 5


@notifications_blueprint.route("/notifications", methods=["GET"])
@require_auth(scopes=["notifications:read"])
def get_auth_user_notifications(auth_user: User) -> Dict:
    """
    Get authenticated user notifications.

    **Scope**: ``notifications:read``

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /api/notifications HTTP/1.1

    - with some query parameters:

    .. sourcecode:: http

      GET /api/notifications?page=2&status=unread  HTTP/1.1

    **Example responses**:

    - returning at least one notification:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "notifications": [
            {
              "created_at": "Wed, 04 Dec 2024 10:06:35 GMT",
              "from": {
                "created_at": "Wed, 04 Dec 2024 09:07:08 GMT",
                "followers": 0,
                "following": 0,
                "follows": "pending",
                "is_followed_by": "false",
                "nb_workouts": 0,
                "picture": true,
                "role": "admin",
                "suspended_at": null,
                "username": "admin"
              },
              "id": "be26f35a-1239-44ff-a012-12fb835fa26c",
              "marked_as_read": false,
              "type": "follow_request"
            }
          ],
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 1,
            "total": 1
          },
          "status": "success"
        }

    - returning no notifications

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "notifications": [],
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 0,
            "total": 0
          },
          "status": "success"
        }

    :query integer page: page if using pagination (default: 1)
    :query string order: sorting order: ``asc``, ``desc`` (default: ``desc``)
    :query string status: notification read status (``read``, ``unread``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    """
    params = request.args.copy()
    page = int(params.get("page", 1))
    order = params.get("order", "desc")
    status = params.get("status")
    marked_as_read = None
    if status == "read":
        marked_as_read = True
    if status == "unread":
        marked_as_read = False
    event_type = params.get("type")

    blocked_users = auth_user.get_blocked_user_ids()
    blocked_by_users = auth_user.get_blocked_by_user_ids()
    following_ids = auth_user.get_following_user_ids()

    filters = [
        Notification.to_user_id == auth_user.id,
        Notification.from_user_id.not_in(blocked_users),
        or_(
            (
                and_(
                    (
                        or_(
                            Notification.event_type != "workout_comment",
                            and_(
                                Notification.event_type == "workout_comment",
                                Notification.from_user_id.not_in(
                                    blocked_by_users
                                ),
                                or_(
                                    Comment.text_visibility
                                    == VisibilityLevel.PUBLIC,
                                    and_(
                                        Comment.text_visibility
                                        == VisibilityLevel.FOLLOWERS,
                                        Notification.from_user_id.in_(
                                            following_ids
                                        ),
                                    ),
                                ),
                            ),
                        )
                    ),
                    User.suspended_at == None,  # noqa
                )
            ),
            (Notification.event_type.in_(["report", "suspension_appeal"])),
        ),
    ]
    if marked_as_read is not None:
        filters.append(Notification.marked_as_read == marked_as_read)
    if event_type:
        filters.append(Notification.event_type == event_type)
    notifications_pagination = (
        Notification.query.join(
            User,
            Notification.from_user_id == User.id,
        )
        .outerjoin(
            Comment,
            Notification.event_object_id == Comment.id,
        )
        .filter(*filters)
        .order_by(
            asc(Notification.created_at)
            if order == "asc"
            else desc(Notification.created_at)
        )
        .paginate(
            page=page, per_page=DEFAULT_NOTIFICATION_PER_PAGE, error_out=False
        )
    )
    notifications = notifications_pagination.items

    return {
        "status": "success",
        "notifications": [
            notification.serialize() for notification in notifications
        ],
        "pagination": {
            "has_next": notifications_pagination.has_next,
            "has_prev": notifications_pagination.has_prev,
            "page": notifications_pagination.page,
            "pages": notifications_pagination.pages,
            "total": notifications_pagination.total,
        },
    }


@notifications_blueprint.route(
    "/notifications/<string:notification_id>", methods=["PATCH"]
)
@require_auth(scopes=["notifications:write"])
def update_user_notifications(
    auth_user: User, notification_id: str
) -> Union[Dict, HttpResponse]:
    """
    Update authenticated user notification read status.

    **Scope**: ``notifications:write``

    **Example request**:

    .. sourcecode:: http

      PATCH /api/notifications/be26f35a-1239-44ff-a012-12fb835fa26c HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

       {
          "notification": {
            "created_at": "Wed, 04 Dec 2024 10:06:35 GMT",
            "from": {
              "created_at": "Wed, 04 Dec 2024 09:07:08 GMT",
              "followers": 0,
              "following": 0,
              "follows": "pending",
              "is_followed_by": "false",
              "nb_workouts": 0,
              "picture": true,
              "role": "admin",
              "suspended_at": null,
              "username": "admin"
            },
            "id": "be26f35a-1239-44ff-a012-12fb835fa26c",
            "marked_as_read": true,
            "type": "follow_request"
          },
          "status": "success"
        }

    :param string notification_id: notification short id

    :<json boolean read_status: notification read status

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``notification not found``
    :statuscode 500:
        - ``error, please try again or contact the administrator``
    """
    notification = Notification.query.filter_by(
        uuid=decode_short_id(notification_id), to_user_id=auth_user.id
    ).first()
    if not notification:
        return NotFoundErrorResponse(
            message=f"notification not found (id: {notification_id})"
        )

    params = request.get_json()
    read_status = params.get("read_status")

    try:
        if read_status is not None:
            notification.marked_as_read = read_status
            db.session.commit()

        return {"status": "success", "notification": notification.serialize()}

    except (exc.IntegrityError, exc.StatementError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@notifications_blueprint.route("/notifications/unread", methods=["GET"])
@require_auth(scopes=["notifications:read"])
def get_status(auth_user: User) -> Dict:
    """
    Get if unread notifications exist for authenticated user.

    **Scope**: ``notifications:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/notifications/unread HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

       {
          "status": "success",
          "unread": false
        }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    """
    unread_notifications = (
        Notification.query.join(
            User,
            Notification.from_user_id == User.id,
        )
        .outerjoin(
            Comment,
            Notification.event_object_id == Comment.id,
        )
        .filter(
            Notification.to_user_id == auth_user.id,
            Notification.from_user_id.not_in(auth_user.get_blocked_user_ids()),
            (
                or_(
                    (
                        and_(
                            (
                                or_(
                                    Notification.event_type
                                    != "workout_comment",
                                    and_(
                                        Notification.event_type
                                        == "workout_comment",
                                        Notification.from_user_id.not_in(
                                            auth_user.get_blocked_by_user_ids()
                                        ),
                                        or_(
                                            Comment.text_visibility
                                            == VisibilityLevel.PUBLIC,
                                            and_(
                                                Comment.text_visibility
                                                == VisibilityLevel.FOLLOWERS,
                                                Notification.from_user_id.in_(
                                                    auth_user.get_following_user_ids()
                                                ),
                                            ),
                                        ),
                                    ),
                                )
                            ),
                            User.suspended_at == None,  # noqa
                        )
                    ),
                    (
                        Notification.event_type.in_(
                            ["report", "suspension_appeal"]
                        )
                    ),
                )
            ),
            Notification.marked_as_read == False,  # noqa
        )
        .count()
    )
    return {
        "status": "success",
        "unread": unread_notifications > 0,
    }


@notifications_blueprint.route(
    "/notifications/mark-all-as-read", methods=["POST"]
)
@require_auth(scopes=["notifications:write"])
def mark_all_as_read(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Mark all authenticated user notifications as read.

    **Scope**: ``notifications:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/notifications/mark-all-as-read HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

       {
          "status": "success"
        }

    :<json boolean type: notification type (optional)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 500:
        - ``error, please try again or contact the administrator``
    """
    params = request.get_json(silent=True)
    event_type = params.get("type") if params else None
    try:
        filters = [
            Notification.to_user_id == auth_user.id,
            Notification.marked_as_read == False,  # noqa
        ]
        if event_type is not None:
            filters.append(Notification.event_type == event_type)
        Notification.query.filter(*filters).update(
            {Notification.marked_as_read: True}, synchronize_session=False
        )
        db.session.commit()
    except Exception as e:
        return handle_error_and_return_response(e, db=db)
    return {"status": "success"}


@notifications_blueprint.route("/notifications/types", methods=["GET"])
@require_auth(scopes=["notifications:read"])
def get_notification_types(auth_user: User) -> Dict:
    """
    Get types of notifications received by authenticated user.

    **Scope**: ``notifications:read``

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /api/notifications/types HTTP/1.1

    - with query parameter:

    .. sourcecode:: http

      GET /api/notifications/types?status=unread  HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

       {
          "notification_types": [
            "mention"
          ],
          "status": "success"
       }

    :query string status: notification read status (``read``, ``unread``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    """
    status = request.args.copy().get("status")
    marked_as_read = None
    if status == "read":
        marked_as_read = True
    if status == "unread":
        marked_as_read = False
    filters = [Notification.to_user_id == auth_user.id]
    if marked_as_read is not None:
        filters.append(Notification.marked_as_read == marked_as_read)
    notification_types = (
        db.session.query(Notification.event_type).filter(*filters).distinct()
    )
    return {
        "notification_types": [
            notification_type[0] for notification_type in notification_types
        ],
        "status": "success",
    }



from datetime import timedelta
from typing import Dict, List, Optional, Tuple, Union

from flask import Blueprint, request
from sqlalchemy import exc, func
from sqlalchemy.dialects.postgresql import insert

from fittrackee import db
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    DataNotFoundErrorResponse,
    ForbiddenErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.users.models import (
    User,
    UserSportPreference,
    UserSportPreferenceEquipment,
)
from fittrackee.utils import decode_short_id
from fittrackee.visibility_levels import VisibilityLevel
from fittrackee.workouts.models import Sport, Workout

from .exceptions import InvalidEquipmentsException
from .models import Equipment, EquipmentType, WorkoutEquipment
from .utils import SPORT_EQUIPMENT_TYPES

equipments_blueprint = Blueprint("equipments", __name__)


def handle_default_sports(
    default_for_sport_ids: List[int],
    auth_user: User,
    equipment_type: EquipmentType,
) -> List[UserSportPreference]:
    user_sport_preferences = []
    for sport_id in default_for_sport_ids:
        sport = Sport.query.filter_by(id=sport_id).first()
        if not sport:
            raise InvalidEquipmentsException(
                f"sport (id {sport_id}) does not exist"
            )

        # check if sport is valid for equipment type
        if (
            equipment_type.label != "Misc"
            and sport.label
            not in SPORT_EQUIPMENT_TYPES.get(equipment_type.label, [])
        ):
            raise InvalidEquipmentsException(
                f"invalid sport '{sport.label}' for equipment "
                f"type '{equipment_type.label}'"
            )

        user_sport_preference = UserSportPreference.query.filter_by(
            user_id=auth_user.id,
            sport_id=sport_id,
        ).first()
        if not user_sport_preference:
            user_sport_preference = UserSportPreference(
                user_id=auth_user.id,
                sport_id=sport_id,
                stopped_speed_threshold=sport.stopped_speed_threshold,
                pace_speed_display=sport.pace_speed_display,
            )
            db.session.add(user_sport_preference)
            db.session.flush()
        user_sport_preferences.append(user_sport_preference)
    return user_sport_preferences


def update_user_sport_preferences_if_exist(
    equipment: "Equipment",
    user_sport_preferences: Optional[List["UserSportPreference"]],
    auth_user: "User",
    default_for_sport_ids: list[int],
) -> None:
    if user_sport_preferences:
        # remove existing pieces of equipment with the same equipment type
        db.session.query(UserSportPreferenceEquipment).filter(
            UserSportPreferenceEquipment.c.user_id == auth_user.id,
            UserSportPreferenceEquipment.c.sport_id.in_(default_for_sport_ids),
            (
                UserSportPreferenceEquipment.c.equipment_type_id
                == equipment.equipment_type_id
            ),
        ).delete()

        db.session.execute(
            insert(UserSportPreferenceEquipment)
            .values(
                [
                    {
                        "equipment_id": equipment.id,
                        "equipment_type_id": equipment.equipment_type_id,
                        "sport_id": sport.sport_id,
                        "user_id": auth_user.id,
                    }
                    for sport in user_sport_preferences
                ]
            )
            .on_conflict_do_nothing()
        )


@equipments_blueprint.route("/equipments", methods=["GET"])
@require_auth(scopes=["equipments:read"], allow_suspended_user=True)
def get_equipments(auth_user: User) -> Dict:
    """
    Get all user equipments.
    Only the equipment owner can see his equipment.

    Suspended user can access this endpoint.

    **Scope**: ``equipments:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/equipments HTTP/1.1
      Content-Type: application/json

    - with some query parameters (get all equipment of type "Shoes")

    .. sourcecode:: http

      GET /api/equipment?equipment_type_id=1  HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "equipments": [
            {
              "creation_date": "Tue, 21 Mar 2023 06:08:06 GMT",
              "default_for_sport_ids": [],
              "description": "The first shoes added to FitTrackee",
              "equipment_type": {
                "id": 1,
                "is_active": true,
                "label": "Shoe"
              },
              "id": "2UkrViYShoAkg8qSUKnUS4",
              "is_active": true,
              "label": "My shoes",
              "total_distance": 0.0,
              "total_duration_in_hours": 0,
              "user_id": 1,
              "workouts_count": 0
          },
          {
              "creation_date": "Tue, 21 Mar 2023 06:08:29 GMT",
              "default_for_sport_ids": [],
              "description": "The second shoes added to FitTrackee",
              "equipment_type": {
                "id": 1,
                "is_active": true,
                "label": "Shoe"
              },
              "id": "2UkrViYShoAkg8qSUKnUS4",
              "is_active": true,
              "label": "My shoes 2",
              "total_distance": 0.0,
              "total_duration_in_hours": 0,
              "user_id": ,
              "workouts_count": 0
              }
            ]
          }
        },
        "status": "success"
      }

    :query integer equipment_type_id: equipment type id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``

    """
    params = request.args.copy()
    type_id = params.get("equipment_type_id", None)

    filters = [Equipment.user_id == auth_user.id]
    if type_id:
        filters.append(Equipment.equipment_type_id == type_id)

    equipments = Equipment.query.filter(*filters).order_by(Equipment.id).all()

    return {
        "status": "success",
        "data": {
            "equipments": [
                equipment.serialize(current_user=auth_user)
                for equipment in equipments
            ]
        },
    }


@equipments_blueprint.route(
    "/equipments/<string:equipment_short_id>", methods=["GET"]
)
@require_auth(scopes=["equipments:read"], allow_suspended_user=True)
def get_equipment_by_id(
    auth_user: User, equipment_short_id: str
) -> Union[Dict, HttpResponse]:
    """
    Get an equipment item.
    Only the equipment owner can see his equipment.

    Suspended user can access this endpoint.

    **Scope**: ``equipments:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/equipments/2UkrViYShoAkg8qSUKnUS4 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - success

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "equipments": [
            {
              "creation_date": "Tue, 21 Mar 2023 06:08:06 GMT",
              "default_for_sport_ids": [],
              "description": "Another piece of equipment",
              "equipment_type": {
                "id": 1,
                "is_active": true,
                "label": "Shoe"
              },
              "id": "2UkrViYShoAkg8qSUKnUS4",
              "is_active": true,
              "label": "Other user Equipment",
              "total_distance": 0.0,
              "total_duration_in_hours": 0,
              "user_id": 2,
              "workouts_count": 0
            }
          ]
        },
        "status": "success"
      }

    - equipment not found

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

      {
        "data": {
          "equipments": []
        },
        "status": "not found"
      }

    :param string equipment_short_id: equipment short id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    :statuscode 404: ``equipment not found``

    """
    filter_args = {
        "uuid": decode_short_id(equipment_short_id),
        "user_id": auth_user.id,
    }
    equipment = Equipment.query.filter_by(**filter_args).first()
    if equipment:
        return {
            "status": "success",
            "data": {
                "equipments": [equipment.serialize(current_user=auth_user)]
            },
        }
    return DataNotFoundErrorResponse("equipments")


@equipments_blueprint.route("/equipments", methods=["POST"])
@require_auth(scopes=["equipments:write"])
def post_equipment(auth_user: User) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Post a new piece of equipment.

    **Scope**: ``equipments:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/equipments HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 201 CREATED
      Content-Type: application/json

      {
        "data": {
          "equipments": [
            {
              "creation_date": "Tue, 21 Mar 2023 06:08:29 GMT",
              "default_for_sport_ids": [],
              "description": null,
              "equipment_type": {
                "id": 1,
                "is_active": true,
                "label": "Shoe"
              },
              "id": "2UkrViYShoAkg8qSUKnUS4",
              "is_active": true,
              "label": "New equipment from API",
              "total_distance": 0.0,
              "total_duration_in_hours": 0,
              "user_id": 1,
              "workouts_count": 0
            }
          ]
        },
        "status": "created"
      }

    :<json string label: a brief (less than 50 characters) label for
        the piece of equipment
    :<json integer equipment_type: the ID for an equipment type (it must be
        active)
    :<json string description: a (perhaps longer) description of the
        equipment (limited to 200 characters, optional)
    :<json boolean is_active: whether or not this equipment is currently
        active (default: ``true``)
    :<json array of integers default_for_sport_ids: the default sport ids
        to use for this equipment, not mandatory
    :<json string visibility: visibility level (``public``, ``followers_only``,
        ``private``), not mandatory (default value: ``private``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 201: equipment created
    :statuscode 400:
        - ``the 'label' and 'equipment_type_id' parameters must be provided``
        - ``equipment already exists with the same label``
        - ``label exceeds 50 characters``
        - ``invalid equipment type id``
        - ``equipment type is inactive``
        - ``sport (id <sport_id>) does not exist``
        - ``invalid sport '<sport_label>' for equipment
          type '<equipment_type_label>'``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``equipment not found``
    :statuscode 500: ``Error during equipment save``
    """
    equipment_data = request.get_json()
    if (
        not equipment_data
        or not equipment_data.get("label")
        or equipment_data.get("equipment_type_id") is None
    ):
        return InvalidPayloadErrorResponse(
            "the 'label' and 'equipment_type_id' parameters must be provided"
        )

    label = equipment_data["label"]
    if len(label) > 50:
        return InvalidPayloadErrorResponse("label exceeds 50 characters")
    equipment_type_id = equipment_data["equipment_type_id"]

    if (
        Equipment.query.filter_by(user_id=auth_user.id, label=label).first()
        is not None
    ):
        return InvalidPayloadErrorResponse(
            "equipment already exists with the same label"
        )

    equipment_type = EquipmentType.query.filter_by(
        id=equipment_type_id
    ).first()
    if not equipment_type:
        return InvalidPayloadErrorResponse("invalid equipment type id")
    if not equipment_type.is_active:
        return InvalidPayloadErrorResponse("equipment type is inactive")

    default_for_sport_ids = equipment_data.get("default_for_sport_ids", [])
    try:
        user_sport_preferences = handle_default_sports(
            default_for_sport_ids, auth_user, equipment_type
        )
    except InvalidEquipmentsException as e:
        return InvalidPayloadErrorResponse(str(e))

    visibility = equipment_data.get("visibility")
    if visibility is not None:
        try:
            VisibilityLevel(visibility)
        except ValueError:
            return InvalidPayloadErrorResponse("invalid visibility")

    try:
        new_equipment = Equipment(
            user_id=auth_user.id,
            label=label,
            equipment_type_id=equipment_type_id,
            is_active=True,
            description=equipment_data.get("description"),
        )
        db.session.add(new_equipment)
        db.session.flush()

        if visibility:
            new_equipment.visibility = visibility

        update_user_sport_preferences_if_exist(
            new_equipment,
            user_sport_preferences,
            auth_user,
            default_for_sport_ids,
        )
        db.session.commit()

        return (
            {
                "status": "created",
                "data": {
                    "equipments": [
                        new_equipment.serialize(current_user=auth_user)
                    ]
                },
            },
            201,
        )

    except (exc.IntegrityError, ValueError) as e:
        return handle_error_and_return_response(
            error=e,
            message="Error during equipment save",
            status="fail",
            db=db,
        )


@equipments_blueprint.route(
    "/equipments/<string:equipment_short_id>", methods=["PATCH"]
)
@require_auth(scopes=["equipments:write"])
def update_equipment(
    auth_user: User, equipment_short_id: str
) -> Union[Dict, HttpResponse]:
    """
    Update a piece of equipment. Allows a user to change one of their
    equipment's label, description, type or active status.

    Changing equipment type will remove all existing workouts associations
    for that piece of equipment and default sports.

    **Scope**: ``equipments:write``

    **Example request**:

    .. sourcecode:: http

      PATCH /api/equipments/QRj7BY6H2iYjSV8sersFgV HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - success

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "equipments": [
              {
                "creation_date": "Tue, 21 Mar 2023 06:28:10 GMT",
                "default_for_sport_ids": [],
                "description": "Change bike to shoes",
                "equipment_type": {
                  "id": 1,
                  "is_active": true,
                  "label": "Shoe"
                },
                "id": "QRj7BY6H2iYjSV8sersFgV",
                "is_active": true,
                "label": "Updated bike",
                "total_distance": 0.0,
                "total_duration_in_hours": 0,
                "user_id": 1,
                "workouts_count": 0
              }
            ]
          },
          "status": "success"
        }

    - equipment not found

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

      {
        "data": {
          "equipments": []
        },
        "status": "not found"
      }

    :param string equipment_short_id: equipment short id

    :<json string label: a brief (less than 50 characters) label for
        the piece of equipment
    :<json int equipment_type_id: the ID for an equipment type (it must be
        active)
    :<json string description: a (perhaps longer) description of the
        equipment (limited to 200 characters, optional)
    :<json boolean is_active: whether or not this equipment is currently
        active (default: ``true``)
    :<json array of integers default_for_sport_ids: the default sport ids
        to use for this equipment
    :<json string visibility: visibility level (``public``, ``followers_only``,
        ``private``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: equipment updated
    :statuscode 400:
        - ``no request data was supplied``
        - ``no valid parameters supplied``
        - ``equipment already exists with the same label``
        - ``label exceeds 50 characters``
        - ``invalid equipment type id``
        - ``equipment type is inactive``
        - ``sport (id <sport_id>) does not exist``
        - ``invalid sport '<sport_label>' for equipment
          type '<equipment_type_label>'``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``equipment not found``
    :statuscode 500: ``Error during equipment update``

    """
    equipment_data = request.get_json()
    if not equipment_data:
        return InvalidPayloadErrorResponse("no request data was supplied")

    if not any(
        e
        in [
            "label",
            "description",
            "equipment_type_id",
            "is_active",
            "default_for_sport_ids",
            "visibility",
        ]
        for e in equipment_data
    ):
        return InvalidPayloadErrorResponse("no valid parameters supplied")

    new_default_for_sport_ids = equipment_data.get(
        "default_for_sport_ids", None
    )
    user_sport_preferences = None
    try:
        equipment = Equipment.query.filter_by(
            uuid=decode_short_id(equipment_short_id), user_id=auth_user.id
        ).first()
        if not equipment:
            return DataNotFoundErrorResponse("equipments")

        check_default_sports = True
        equipment_type = equipment.equipment_type

        # set new values if they were in the request
        if "is_active" in equipment_data:
            equipment.is_active = equipment_data.get("is_active")
        if "label" in equipment_data:
            label = equipment_data.get("label")
            if len(label) > 50:
                return InvalidPayloadErrorResponse(
                    "label exceeds 50 characters"
                )
            if (
                Equipment.query.filter(
                    Equipment.user_id == auth_user.id,
                    Equipment.label == label,
                    Equipment.id != equipment.id,
                ).first()
                is not None
            ):
                return InvalidPayloadErrorResponse(
                    "equipment already exists with the same label"
                )
            equipment.label = label
        if "description" in equipment_data:
            equipment.description = equipment_data.get("description")

        if "visibility" in equipment_data:
            visibility = equipment_data["visibility"]
            try:
                VisibilityLevel(visibility)
            except ValueError:
                return InvalidPayloadErrorResponse("invalid visibility")
            equipment.visibility = visibility

        if "equipment_type_id" in equipment_data:
            equipment_type_id = equipment_data.get("equipment_type_id")
            equipment_type = EquipmentType.query.filter_by(
                id=equipment_type_id
            ).first()
            if not equipment_type:
                return InvalidPayloadErrorResponse("invalid equipment type id")
            if equipment_type.id != equipment.equipment_type_id:
                if not equipment_type.is_active:
                    return InvalidPayloadErrorResponse(
                        "equipment type is inactive"
                    )

                # remove workouts association on type change
                db.session.query(WorkoutEquipment).filter(
                    WorkoutEquipment.c.equipment_id == equipment.id
                ).delete()
                equipment.total_distance = 0.0
                equipment.total_duration = timedelta()
                equipment.total_moving = timedelta()
                equipment.total_workouts = 0

                # remove default sports
                db.session.query(UserSportPreferenceEquipment).filter(
                    UserSportPreferenceEquipment.c.user_id == auth_user.id,
                    UserSportPreferenceEquipment.c.equipment_id
                    == equipment.id,
                ).all()

                # with changes on equipments but no changes on default sports,
                # the default sports are removed.
                if new_default_for_sport_ids is None or (
                    new_default_for_sport_ids
                    == [
                        sport_preference.sport_id
                        for sport_preference in equipment.default_for_sports
                    ]
                ):
                    new_default_for_sport_ids = []
                    check_default_sports = False

            equipment.equipment_type_id = equipment_type_id

        if check_default_sports and new_default_for_sport_ids is not None:
            try:
                user_sport_preferences = handle_default_sports(
                    new_default_for_sport_ids,
                    auth_user,
                    equipment_type,
                )
            except InvalidEquipmentsException as e:
                return InvalidPayloadErrorResponse(str(e))

        if new_default_for_sport_ids is not None:
            existing_sports = (
                db.session.query(UserSportPreferenceEquipment)
                .filter(
                    UserSportPreferenceEquipment.c.user_id == auth_user.id,
                    UserSportPreferenceEquipment.c.equipment_id
                    == equipment.id,
                )
                .all()
            )

            sport_ids_to_remove = {s[1] for s in existing_sports} - set(
                new_default_for_sport_ids
            )

            if sport_ids_to_remove:
                db.session.query(UserSportPreferenceEquipment).filter(
                    UserSportPreferenceEquipment.c.user_id == auth_user.id,
                    (
                        UserSportPreferenceEquipment.c.equipment_id
                        == equipment.id
                    ),
                    UserSportPreferenceEquipment.c.sport_id.in_(
                        sport_ids_to_remove
                    ),
                ).delete()

            update_user_sport_preferences_if_exist(
                equipment,
                user_sport_preferences,
                auth_user,
                new_default_for_sport_ids,
            )
        db.session.commit()

        return {
            "status": "success",
            "data": {
                "equipments": [equipment.serialize(current_user=auth_user)]
            },
        }

    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(
            error=e,
            db=db,
            message="Error during equipment update",
            status="fail",
        )


@equipments_blueprint.route(
    "/equipments/<string:equipment_short_id>/refresh", methods=["POST"]
)
@require_auth(scopes=["equipments:write"])
def refresh_equipment(
    auth_user: User, equipment_short_id: str
) -> Union[Dict, HttpResponse]:
    """
    Refresh equipment totals (in case values are incorrect).

    **Scope**: ``equipments:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/equipments/QRj7BY6H2iYjSV8sersFgV/refresh HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "equipments": [
              {
                "creation_date": "Tue, 21 Mar 2023 06:28:10 GMT",
                "default_for_sport_ids": [],
                "description": "My Shoes",
                "equipment_type": {
                  "id": 1,
                  "is_active": true,
                  "label": "Shoe"
                },
                "id": "QRj7BY6H2iYjSV8sersFgV",
                "is_active": true,
                "label": "Updated bike",
                "total_distance": 6.0,
                "total_duration_in_hours": 1,
                "user_id": 1,
                "workouts_count": 1
              }
            ]
          },
          "status": "success"
        }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: equipment updated
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``equipment not found``
    :statuscode 500: ``Error during equipment save``
    """
    equipment = Equipment.query.filter_by(
        uuid=decode_short_id(equipment_short_id), user_id=auth_user.id
    ).first()
    if not equipment:
        return DataNotFoundErrorResponse("equipments")

    try:
        totals = (
            db.session.query(
                func.sum(Workout.distance).label("total_distance"),
                func.sum(Workout.duration).label("total_duration"),
                func.sum(Workout.moving).label("total_moving"),
                func.count(Workout.id).label("total_workouts"),
            )
            .join(WorkoutEquipment)
            .filter(WorkoutEquipment.c.equipment_id == equipment.id)
            .one()
            ._asdict()
        )
        equipment.total_distance = (
            0.0
            if totals["total_distance"] is None
            else totals["total_distance"]
        )
        equipment.total_duration = (
            timedelta()
            if totals["total_duration"] is None
            else totals["total_duration"]
        )
        equipment.total_moving = (
            timedelta()
            if totals["total_moving"] is None
            else totals["total_moving"]
        )
        equipment.total_workouts = totals["total_workouts"]

        db.session.commit()

        return {
            "status": "success",
            "data": {
                "equipments": [equipment.serialize(current_user=auth_user)]
            },
        }

    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(
            error=e,
            db=db,
            message="Error during equipment update",
            status="fail",
        )


@equipments_blueprint.route(
    "/equipments/<string:equipment_short_id>", methods=["DELETE"]
)
@require_auth(scopes=["equipments:write"])
def delete_equipment(
    auth_user: User, equipment_short_id: str
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Delete a piece of equipment.

    A user can only delete their own equipment, and
    only if there are no workouts associated with that equipment (unless
    forced). If equipment was associated with any workouts and deletion is
    forced, the association between this equipment and those workouts will
    be removed. If this equipment was a default for any sport, that default
    will be removed (set to NULL).

    **Scope**: ``equipments:write``

    **Example request**:

    .. sourcecode:: http

      DELETE /api/equipments/QRj7BY6H2iYjSV8sersFgV HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 204 NO CONTENT
      Content-Type: application/json

    :param string equipment_short_id: equipment short id

    :query force: if supplied as argument (no value required), will force
                  deletion of the equipment and remove that equipment
                  from associated workouts

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 204: equipment deleted
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
        - ``you cannot delete equipment that has workouts associated with it
          without 'force' parameter``
    :statuscode 404: ``equipment not found``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    force_delete = "force" in request.args
    equipment_uuid = decode_short_id(equipment_short_id)

    try:
        equipment = Equipment.query.filter_by(
            uuid=equipment_uuid, user_id=auth_user.id
        ).first()
        if not equipment:
            return DataNotFoundErrorResponse("equipments")
        if equipment.total_workouts > 0 and not force_delete:
            return ForbiddenErrorResponse(
                f"Cannot delete equipment that has associated workouts. "
                f"Equipment id {equipment.short_id} has "
                f"{equipment.total_workouts} associated "
                f"workout{'' if equipment.total_workouts == 1 else 's'}. "
                f"(Provide argument 'force' as a query parameter to "
                f"override this check)"
            )
        # other condition for deleting attachment to workouts
        # is handled by database cascading

        # NULLing of user sport preferences handled by database cascading

        # delete equipment row
        db.session.query(Equipment).filter(
            Equipment.id == equipment.id
        ).delete()
        db.session.commit()
        return {"status": "no content"}, 204
    except (
        exc.IntegrityError,
        exc.OperationalError,
        ValueError,
        OSError,
    ) as e:  # pragma: no cover
        return handle_error_and_return_response(e, db=db)



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Optional

from fittrackee.responses import ForbiddenErrorResponse, NotFoundErrorResponse

from .exceptions import CommentForbiddenException
from .utils import get_comment

if TYPE_CHECKING:
    from fittrackee.users.models import User


def check_workout_comment(only_owner: bool = True) -> Callable:
    def decorator_check_workout_comment(f: Callable) -> Callable:
        @wraps(f)
        def wrapper_check_workout_comment(
            *args: Any, **kwargs: Any
        ) -> Callable:
            auth_user: Optional["User"] = args[0]
            comment_short_id: str = kwargs["comment_short_id"]

            try:
                comment = get_comment(comment_short_id, auth_user)
            except CommentForbiddenException:
                return NotFoundErrorResponse(
                    f"workout comment not found (id: {comment_short_id})"
                )
            if only_owner and (
                not auth_user or auth_user.id != comment.user.id
            ):
                return ForbiddenErrorResponse()
            return f(auth_user, comment)

        return wrapper_check_workout_comment

    return decorator_check_workout_comment
